import statistics

import pandas as pd
import seaborn as sns

from functools import reduce

from mongoengine import DoesNotExist
from pycoshark.mongomodels import Project, VCSSystem, TestState, Tag, Mutation, CommitChanges, File
from matplotlib import pyplot as plt
from matplotlib_venn import venn2
import os
from scipy import stats
from tabulate import tabulate

from emse_2018 import analyzed_java_versions, test_type_metric, get_cov_test_lines, merge_parameterized_tests, \
    get_test_states, merge, get_killed_seeded_defects, get_disjoint_mutant_set


def get_classification_of_seeded_defect(seeded_defect):
    defect_classification = {
        'COMPUTATION': False,
        'DATA': False,
        'INTERFACE': False,
        'LOGIC/CONTROL': False,
        'OTHER': False
    }

    for commit_id in seeded_defect.commit_ids:
        try:
            commit_change = CommitChanges.objects(new_commit_id=commit_id).get()
            for file_id, classification in commit_change.classification.items():
                file_path = File.objects(id=file_id).only('path').get().path

                # Skip test changes, as they are not implemented!
                if file_path.startswith("src/test") or "test" in file_path.split("/")[-1].lower() \
                        or "validate" in file_path.split("/")[-1].lower():
                    continue

                # If the classification was found in the change, we say the defect belongs to this class
                for defect_class, classification_number in classification.items():
                    defect_classification[defect_class] = True
        except DoesNotExist:
            print("Commit changes for commit with id %s does not exist!" % commit_id)

    return defect_classification


def plot_venn_diagram(title, data1, data2, data3, overall, axes):
    #grey_box_text = "Overall: %s" % len(overall)
    #plt.text(-0.9, 0.85, grey_box_text, bbox={'facecolor': 'grey', 'alpha': 0.5, 'pad': 10}, axes=axes)
    axes.set_title(title)

    v = venn2(subsets=(len(data1),
                       len(data2),
                       len(data3)),
              set_labels=('UT', 'IT'),
              ax=axes,
              )

    try:
        v.get_patch_by_id('100').set_color("#0868ac")
    except AttributeError:
        pass

    try:
        v.get_patch_by_id('010').set_color("#bae4bc")
    except AttributeError:
        pass

    try:
        v.get_patch_by_id('110').set_color("#7bccc4")
    except AttributeError:
        pass


def create_plots(results, ut_tline_sum, it_tlines_sum, name):
    # Creates venn diagram plots

    all_killed_fig, all_killed_axes = plt.subplots(3, 2, figsize=(6, 10))
    all_killed_fig.suptitle(name)

    n_row = 0
    n_col = -1
    for change_type, dictionary in sorted(results.items()):
        n_col += 1
        if n_col > 1:
            n_row += 1
            n_col = 0

        killed_by_unit_tests_only = dictionary['killed_by_unit_test'] - dictionary['killed_by_both']
        killed_by_integration_tests_only = dictionary['killed_by_integration_test'] - dictionary['killed_by_both']
        killed_axes = all_killed_axes[n_row][n_col]

        ut_score = (len(killed_by_unit_tests_only) / ut_tline_sum)
        it_score = (len(killed_by_integration_tests_only) / it_tlines_sum)

        title = "%s - %.3f / %.3f" % (change_type, ut_score, it_score)

        plot_venn_diagram(title,
                          killed_by_unit_tests_only,
                          killed_by_integration_tests_only,
                          dictionary['killed_by_both'],
                          mutations,
                          killed_axes)

    all_killed_fig.show()
    all_killed_fig.savefig(os.path.join(os.path.dirname(__file__), "figures", name.replace(" ", "-")+".pdf"))


def create_seeded_results(project_id, commit_id):
    # Creates seeded results based on change class

    seeded_killed_by_unit_test, seeded_killed_by_integration_test, seeded_killed_by_both = get_killed_seeded_defects(
        project_id, commit_id)

    seeded_results = {
        'ALL': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'DATA': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'COMPUTATION': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([]),
        },
        'INTERFACE': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'LOGIC/CONTROL': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([]),
        },
        'OTHER': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        }
    }

    for seeded_defect in seeded_killed_by_unit_test:
        for classification, result in get_classification_of_seeded_defect(seeded_defect).items():
            if result:
                seeded_results[classification]['killed_by_unit_test'].add(seeded_defect)
                seeded_results['ALL']['killed_by_unit_test'].add(seeded_defect)

    for seeded_defect in seeded_killed_by_integration_test:
        for classification, result in get_classification_of_seeded_defect(seeded_defect).items():
            if result:
                seeded_results[classification]['killed_by_integration_test'].add(seeded_defect)
                seeded_results['ALL']['killed_by_integration_test'].add(seeded_defect)

    for seeded_defect in seeded_killed_by_both:
        for classification, result in get_classification_of_seeded_defect(seeded_defect).items():
            if result:
                seeded_results[classification]['killed_by_both'].add(seeded_defect)
                seeded_results['ALL']['killed_by_both'].add(seeded_defect)

    return seeded_results


def create_mutation_results(mutations, disjoint_mutants=set()):
    # Create mutation results based on change class

    mutation_results = {
        'ALL': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'DATA': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'COMPUTATION': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([]),
        },
        'INTERFACE': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        },
        'LOGIC/CONTROL': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([]),
        },
        'OTHER': {
            'killed_by_unit_test': set([]),
            'killed_by_integration_test': set([]),
            'killed_by_both': set([])
        }
    }

    for mutation_id, test_type_dict in mutations.items():
        if disjoint_mutants and mutation_id not in disjoint_mutants:
            continue

        classification = change_types[mutation_id]
        store_to = mutation_results.get(classification)

        if len(test_type_dict['UNIT_KILLED']) > 0:
            mutation_results['ALL']['killed_by_unit_test'].add(mutation_id)
            store_to['killed_by_unit_test'].add(mutation_id)

        if len(test_type_dict['INTEGRATION_KILLED']) > 0:
            mutation_results['ALL']['killed_by_integration_test'].add(mutation_id)
            store_to['killed_by_integration_test'].add(mutation_id)

        if len(test_type_dict['INTEGRATION_KILLED']) > 0 and len(test_type_dict['UNIT_KILLED']) > 0:
            mutation_results['ALL']['killed_by_both'].add(mutation_id)
            store_to['killed_by_both'].add(mutation_id)

    return mutation_results


def handle_results(df, results, data_set, project_name, ut_testlc_sum, it_testlc_sum):
    for change_type, result in results.items():
        df.loc[len(df)] = [project_name, 'Unit', data_set, change_type,
                           len(result['killed_by_unit_test']-result['killed_by_both']),
                           len(result['killed_by_unit_test']-result['killed_by_both']) / (ut_testlc_sum / 1000)]
        df.loc[len(df)] = [project_name, 'Integration', data_set, change_type,
                           len(result['killed_by_integration_test']-result['killed_by_both']),
                           len(result['killed_by_integration_test']-result['killed_by_both']) / (it_testlc_sum / 1000)]
        df.loc[len(df)] = [project_name, 'Both', data_set, change_type,
                           len(result['killed_by_both']),
                           len(result['killed_by_both']) / ((ut_testlc_sum+it_testlc_sum) / 1000)]
    return df


defect_types = ['COMPUTATION', 'DATA', 'INTERFACE', 'LOGIC/CONTROL']
analyzed_data_sets = ['ALL', 'DISJ']
table_mutation = []
table_disj_mutation = []

df = pd.DataFrame(columns=['Project', 'Test Level', 'Data Set', 'Defect Type', 'Killed', 'Score'])

# Go through all analyzed versions
for project_information in analyzed_java_versions:
    project_name = project_information[0]
    project_version = project_information[1]

    print("Looking at project %s (%s)..." % (project_name, project_version))

    project_id = Project.objects(name=project_name).get().id
    vcs_system = VCSSystem.objects(project_id=project_id).get()
    commit_id = Tag.objects(vcs_system_id=vcs_system.id, name=project_version).get().commit_id

    # Get all test states that have a mutation_res array
    unit_test_states = list(TestState.objects(**{'mutation_res__ne': None,
                                            'commit_id': commit_id,
                                            'metrics__'+test_type_metric: 'UNIT'})
                            .only('id', 'name', 'metrics').all())
    unit_test_states_ids = [state.id for state in merge_parameterized_tests(unit_test_states)]

    integration_test_states = TestState.objects(**{'mutation_res__ne': None,
                                                   'commit_id': commit_id,
                                                   'metrics__'+test_type_metric: 'INTEGRATION'})\
        .only('id', 'name', 'metrics').all()
    integration_test_states_ids = [state.id for state in merge_parameterized_tests(integration_test_states)]

    # Get TestLC for unit and integration tests
    ut_tline_sum = get_cov_test_lines(unit_test_states, unit_test_states_ids)
    it_tlines_sum = get_cov_test_lines(integration_test_states, integration_test_states_ids)

    print("Analyzing %d unit tests and %d integration tests..." % (len(unit_test_states), len(integration_test_states)))

    # Get all mutations killed by unit and integration tests
    mutations = {}
    unit_killed = get_test_states(unit_test_states_ids, 'UNIT', 'KILLED')
    it_killed = get_test_states(integration_test_states_ids, 'INTEGRATION', 'KILLED')
    reduce(merge, [mutations, unit_killed, it_killed])

    # Get all change types for all mutations
    change_types = {}
    for mutation in Mutation.objects(id__in=mutations.keys()).only('id', 'classification').all():
        change_types[mutation.id] = mutation.classification

    # Get Results
    mutation_results = create_mutation_results(mutations)
    df = handle_results(df, mutation_results, 'ALL', project_name, ut_tline_sum, it_tlines_sum)
    create_plots(mutation_results, ut_tline_sum, it_tlines_sum, "Mutation Data (%s)" % project_name)

    disjoint_mutants_results = create_mutation_results(mutations, get_disjoint_mutant_set(mutations))
    df = handle_results(df, disjoint_mutants_results, 'DISJ', project_name, ut_tline_sum, it_tlines_sum)
    create_plots(disjoint_mutants_results, ut_tline_sum, it_tlines_sum, "Disjoint Mutation Data (%s)" % project_name)


    table_mutation.append([project_name,
                           "%.2f" % (len(mutation_results['COMPUTATION']['killed_by_unit_test']-mutation_results['COMPUTATION']['killed_by_both'])/(ut_tline_sum/1000)),
                           "%.2f" % (len(mutation_results['COMPUTATION']['killed_by_integration_test']-mutation_results['COMPUTATION']['killed_by_both'])/(it_tlines_sum/1000)),
                           "%.2f" % (len(mutation_results['DATA']['killed_by_unit_test']-mutation_results['DATA']['killed_by_both'])/(ut_tline_sum/1000)),
                           "%.2f" % (len(mutation_results['DATA']['killed_by_integration_test']-mutation_results['DATA']['killed_by_both'])/(it_tlines_sum/1000)),
                           "%.2f" % (len(mutation_results['INTERFACE']['killed_by_unit_test']-mutation_results['INTERFACE']['killed_by_both'])/(ut_tline_sum/1000)),
                           "%.2f" % (len(mutation_results['INTERFACE']['killed_by_integration_test']-mutation_results['INTERFACE']['killed_by_both'])/(it_tlines_sum/1000)),
                           "%.2f" % (len(mutation_results['LOGIC/CONTROL']['killed_by_unit_test']-mutation_results['LOGIC/CONTROL']['killed_by_both'])/(ut_tline_sum/1000)),
                           "%.2f" % (len(mutation_results['LOGIC/CONTROL']['killed_by_integration_test']-mutation_results['LOGIC/CONTROL']['killed_by_both'])/(it_tlines_sum/1000))
                           ])

    table_disj_mutation.append([project_name,
                               "%.3f" % (len(disjoint_mutants_results['COMPUTATION']['killed_by_unit_test']-disjoint_mutants_results['COMPUTATION']['killed_by_both'])/(ut_tline_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['COMPUTATION']['killed_by_integration_test']-disjoint_mutants_results['COMPUTATION']['killed_by_both'])/(it_tlines_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['DATA']['killed_by_unit_test']-disjoint_mutants_results['DATA']['killed_by_both'])/(ut_tline_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['DATA']['killed_by_integration_test']-disjoint_mutants_results['DATA']['killed_by_both'])/(it_tlines_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['INTERFACE']['killed_by_unit_test']-disjoint_mutants_results['INTERFACE']['killed_by_both'])/(ut_tline_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['INTERFACE']['killed_by_integration_test']-disjoint_mutants_results['INTERFACE']['killed_by_both'])/(it_tlines_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['LOGIC/CONTROL']['killed_by_unit_test']-disjoint_mutants_results['LOGIC/CONTROL']['killed_by_both'])/(ut_tline_sum/1000)),
                               "%.3f" % (len(disjoint_mutants_results['LOGIC/CONTROL']['killed_by_integration_test']-disjoint_mutants_results['LOGIC/CONTROL']['killed_by_both'])/(it_tlines_sum/1000))
                           ])


data_set_with_table_handlers = {
    'ALL': table_mutation,
    'DISJ': table_disj_mutation
}

test_levels = ['Unit', 'Integration']

for data_set, table_handler in data_set_with_table_handlers.items():
    # Visualization
    sns.boxplot(x="Test Level", y="Score", hue="Defect Type",
                hue_order=defect_types,
                data=df.loc[(df['Data Set'] == data_set) & (df['Defect Type'].isin(defect_types))
                            & (df['Test Level'].isin(test_levels))])\
        .set_title("Data Set: "+data_set)
    ax = sns.swarmplot(x="Test Level", y="Score", hue="Defect Type",
                       hue_order=defect_types,
                       data=df.loc[(df['Data Set'] == data_set) & (df['Defect Type'].isin(defect_types)) &
                                                       (df['Test Level'].isin(test_levels))],
                       color=".25", dodge=True)
    # Remove one legend
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles[:4], labels[:4])
    plt.savefig(os.path.join(os.path.dirname(__file__), "figures", "%s_detailed.pdf" % data_set))
    plt.show()

    # Table
    print("Table for %s data set" % data_set)
    table_handler.append(['Mean',
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'COMPUTATION'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'COMPUTATION'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'DATA'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'DATA'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'INTERFACE'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'INTERFACE'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'LOGIC/CONTROL'), 'Score']),
                           "%.2f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'LOGIC/CONTROL'), 'Score'])])

    table_handler.append(['StDev',
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'COMPUTATION'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'COMPUTATION'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'DATA'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'DATA'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'INTERFACE'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'INTERFACE'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit') & (
                                       df['Defect Type'] == 'LOGIC/CONTROL'), 'Score']),
                           "%.2f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration') & (
                                       df['Defect Type'] == 'LOGIC/CONTROL'), 'Score'])])
    print(tabulate(table_handler, tablefmt="latex"))


    # Statistical Tests
    for defect_type in defect_types:
        print("Results for data set %s and defect type %s" % (data_set, defect_type))
        print(stats.shapiro(
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Unit') &
                   (df['Defect Type'] == defect_type), 'Score'])
        )
        print(stats.shapiro(
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Integration') &
                   (df['Defect Type'] == defect_type), 'Score'])
        )
        print(stats.levene(
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Unit') &
                   (df['Defect Type'] == defect_type), 'Score'],
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Integration') &
                   (df['Defect Type'] == defect_type), 'Score'])
        )
        print(stats.mannwhitneyu(
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Unit') &
                   (df['Defect Type'] == defect_type), 'Score'],
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Integration') & (
                    df['Defect Type'] == defect_type), 'Score'])
        )
        print(stats.ttest_ind(
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Unit') &
                   (df['Defect Type'] == defect_type), 'Score'],
            df.loc[(df['Data Set'] == data_set) &
                   (df['Test Level'] == 'Integration') & (
                    df['Defect Type'] == defect_type), 'Score'])
        )
