import seaborn as sns
from matplotlib import pyplot as plt
from pycoshark.utils import create_mongodb_uri_string
from mongoengine import connect

from collections import defaultdict

from pycoshark.mongomodels import TestState

from mongoengine import Document, ObjectIdField, ListField, BooleanField, StringField



plt.style.use(u'ggplot')
color_blind_and_printer_friendly = ["#0868ac", "#43a2ca", "#7bccc4", "#bae4bc"]
sns.set_palette(sns.color_palette(color_blind_and_printer_friendly))
test_type_metric = 'cov_ieee_met'

# Connect to database

# LOCAL
user = None
password = None
host = 'localhost'
port = '27017'
authentication_db = None
database = "comfort"
ssl_enabled = None

uri = create_mongodb_uri_string(user, password, host, port, authentication_db, ssl_enabled)
connect(database, host=uri)

# Analyzed versions of project
analyzed_java_versions = [
   ('commons-beanutils','BEANUTILS_1_9_3'),
   ('commons-codec', 'commons-codec-1.11'),
   ('commons-collections', 'collections-4.1'),
   ('commons-io', '2.5'),
   ('commons-lang', 'LANG_3_7'),
   ('commons-math', 'MATH_3_6'),
   ('druid', '1.1.0'),
   ('fastjson', '1.2.41'),
   ('google-gson', 'gson-parent-2.8.0'),
   ('guice', '4.1'),
   ('HikariCP', 'HikariCP-2.7.0'),
   ('jackson-core', 'jackson-core-2.9.0'),
   ('jfreechart', 'v1.5.0'),
   ('joda-time', 'v2.9'),
   ('jsoup', 'jsoup-1.11.1'),
   ('mybatis-3', 'mybatis-3.4.0'),
   ('zxing', 'zxing-3.3.0')
]

class DetectedDefect(Document):
    project_id = ObjectIdField()
    release_commit_id = ObjectIdField()
    issue_ids = ListField(ObjectIdField())
    commit_ids = ListField(ObjectIdField())
    is_integrateable = BooleanField()
    failed_test_ids = ListField(ObjectIdField())
    errored_test_ids = ListField(ObjectIdField())
    falsely_errored_test_ids = ListField(ObjectIdField())
    comment = StringField()

    def __str__(self):
        return "<project_id: %s>, <release_commit_id: %s>, <issue_ids: %s>, <commit_ids: %s>, " \
               "<is_integrateable: %s>, <failed_test_ids: %s>, <errored_test_ids: %s>, <falsely_errored_test_ids: %s>, " \
               "<comment: %s>" % (
            self.project_id,
            self.release_commit_id,
            self.issue_ids,
            self.commit_ids,
            self.is_integrateable,
            self.failed_test_ids,
            self.errored_test_ids,
            self.falsely_errored_test_ids,
            self.comment
        )


def get_killed_seeded_defects(project_id, commit_id):
    # Get all test states that have a mutation_res array
    unit_test_states = list(TestState.objects(**{'mutation_res__ne': None,
                                                 'commit_id': commit_id,
                                                 'metrics__' + test_type_metric: 'UNIT'})
                            .only('id', 'name', 'metrics').all())

    unit_test_states_ids = [state.id for state in unit_test_states]

    integration_test_states = TestState.objects(**{'mutation_res__ne': None,
                                                   'commit_id': commit_id,
                                                   'metrics__' + test_type_metric: 'INTEGRATION'}) \
        .only('id', 'name', 'metrics').all()

    integration_test_states_ids = [state.id for state in integration_test_states]

    killed_by_unit_test = set([])
    killed_by_integration_test = set([])
    killed_by_both = set([])

    for seeded_defect in DetectedDefect.objects(project_id=project_id, release_commit_id=commit_id,
                                                is_integrateable=True).all():
        for failed_test in seeded_defect.failed_test_ids + seeded_defect.errored_test_ids:
            if failed_test in unit_test_states_ids and failed_test in integration_test_states_ids:
                killed_by_both.add(seeded_defect)
                continue

            if failed_test in unit_test_states_ids:
                killed_by_unit_test.add(seeded_defect)

            if failed_test in integration_test_states_ids:
                killed_by_integration_test.add(seeded_defect)

    print("Seeded Defects killed (Unit Test): %d" % len(killed_by_unit_test))
    print("Seeded Defects killed (Integration Test): %d" % len(killed_by_integration_test))
    print("Seeded Defects killed (Both): %d" % len(killed_by_both))

    return killed_by_unit_test, killed_by_integration_test, killed_by_both


def get_disjoint_mutant_set(mutations):
    new_mutations = {}
    for mutation_id, test_type_dict in mutations.items():
        new_mutations[mutation_id] = set()

        for key, value in test_type_dict.items():
            new_mutations[mutation_id].update(value)

    all_mutations = list(new_mutations.keys())
    final_mutations = set()
    while len(all_mutations) > 0:
        max_subsumed = 0
        subsumed_mut = set()
        max_mut_subsuming = None

        for mutation in all_mutations:
            sub_m = set()

            for mutation2 in all_mutations:
                if new_mutations[mutation2].issubset(new_mutations[mutation]):
                    sub_m.add(mutation2)

            if len(sub_m) > max_subsumed:
                max_subsumed = len(sub_m)
                max_mut_subsuming = mutation
                subsumed_mut = sub_m

        final_mutations.add(max_mut_subsuming)

        for mutation_id in subsumed_mut:
            all_mutations.remove(mutation_id)

    return final_mutations


# Getting test states
# Result: <mutation_id>: [TestState, TestState, ...]
# Where the TestStates have an id which is in ids and a status which is in status
def get_test_states(ids, test_type, status):

    pipeline = [
        {
            "$match": {"_id": {'$in': ids}}
        },
        {
            "$unwind": '$mutation_res'
        },
        {
            "$match": {
                'mutation_res.result': status
            }
        },
        {
            "$group": {
                "_id": "$mutation_res.mutation_id",
                "test_states": {"$addToSet": '$_id'}
            },
        }
    ]

    mutations = defaultdict(lambda: defaultdict(set))
    all_mutations = list(TestState.objects.aggregate(*pipeline, allowDiskUse=True))
    for mutation in all_mutations:
        mutations[mutation['_id']].get(test_type + "_" + status, dict())

        # For the no_coverage calculation we do not need the test states, the length is totally sufficient
        if status == 'NO_COVERAGE':
            mutations[mutation['_id']][test_type + "_" + status] = len(mutation['test_states'])
        else:
            mutations[mutation['_id']][test_type+"_"+status] = set(mutation['test_states'])
    return mutations


def merge_parameterized_tests(tests):
    final_tests = []
    tests_to_skip = []
    for test in tests:
        # We can just skip it here, as the mutation data collection is done for all parameterized tests together anyway
        # So we do not need to merge the results!
        if test.name.split("[")[0] in tests_to_skip:
            continue

        if "[" in test.name:
            tests_to_skip.append(test.name.split("[")[0])

        final_tests.append(test)
    return final_tests


def get_cov_test_lines(test_states, test_ids):
    testlc_sum = 0
    for test_state in test_states:
        if test_state.id not in test_ids:
            continue

        testlc_sum += int(test_state.metrics['cov_tlines'])

    return testlc_sum


# Helper function to merge dicts into each other
def merge(a, b, path=None):
    "merges b into a"
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a




