import statistics

import pandas as pd
import seaborn as sns
import timeit
import os
from functools import reduce
from pycoshark.mongomodels import Project, VCSSystem, Tag, TestState
from matplotlib import pyplot as plt
from scipy import stats
from tabulate import tabulate

from emse_2018 import analyzed_java_versions, test_type_metric, merge_parameterized_tests, get_cov_test_lines, get_test_states, merge, \
    get_killed_seeded_defects, get_disjoint_mutant_set


def get_number_of_killed_mutations(mutations, disjoint_mutants=set()):
    killed_by_both = 0
    killed_by_unit_test = 0
    killed_by_integration_test = 0

    for mutation_id, test_type_dict in mutations.items():
        # Filter out mutants
        if disjoint_mutants and mutation_id not in disjoint_mutants:
            continue

        if len(test_type_dict['UNIT_KILLED']) > 0 and len(test_type_dict['INTEGRATION_KILLED']) > 0:
            killed_by_both += 1
            continue

        if len(test_type_dict['UNIT_KILLED']) > 0:
            killed_by_unit_test += 1

        if len(test_type_dict['INTEGRATION_KILLED']) > 0:
            killed_by_integration_test += 1

    print("Number of Mutations killed (Unit Test): %d" % killed_by_unit_test)
    print("Number of Mutations killed (Integration Test): %d" % killed_by_integration_test)
    print("Number of Mutations killed (Both): %d" % killed_by_both)

    return killed_by_unit_test, killed_by_integration_test, killed_by_both


start_time = timeit.default_timer()

df = pd.DataFrame(columns=['Project', 'Data Set', 'Test Level', 'Killed', 'Score'])
table_seeded = []
table_mutation = []
table_disj_mutation = []
for project_information in analyzed_java_versions:
    project_name = project_information[0]
    project_version = project_information[1]

    print("Analyzing %s (%s)..." % (project_name, project_version))

    project_id = Project.objects(name=project_name).get().id
    vcs_system = VCSSystem.objects(project_id=project_id).get()
    commit_id = Tag.objects(vcs_system_id=vcs_system.id, name=project_version).get().commit_id

    # Get all test states that have a mutation_res array
    unit_test_states = list(TestState.objects(**{'mutation_res__ne': None,
                                                 'commit_id': commit_id,
                                                 'metrics__' + test_type_metric: 'UNIT'})
                            .only('id', 'name', 'metrics').all())

    # Merge parameterized tests together, so that it is seen as one test
    unit_test_states_ids = [state.id for state in merge_parameterized_tests(unit_test_states)]

    integration_test_states = TestState.objects(**{'mutation_res__ne': None,
                                                   'commit_id': commit_id,
                                                   'metrics__' + test_type_metric: 'INTEGRATION'}) \
        .only('id', 'name', 'metrics').all()

    integration_test_states_ids = [state.id for state in merge_parameterized_tests(integration_test_states)]

    # Sum up TestLC
    ut_testlc_sum = get_cov_test_lines(unit_test_states, unit_test_states_ids)
    it_testlc_sum = get_cov_test_lines(integration_test_states, integration_test_states_ids)

    print("Analyzing %d unit tests and %d integration tests..." % (len(unit_test_states), len(integration_test_states)))

    # Get all mutatations together with the test states that kill these mutations
    # Form: <mutation_id>: [TestStateId, TestStateId, TestStateId]
    mutations = {}
    unit_killed = get_test_states(unit_test_states_ids, 'UNIT', 'KILLED')
    it_killed = get_test_states(integration_test_states_ids, 'INTEGRATION', 'KILLED')
    reduce(merge, [mutations, unit_killed, it_killed])

    # Seeded Data Set
    seeded_killed_by_unit_test, seeded_killed_by_integration_test, seeded_killed_by_both =\
        get_killed_seeded_defects(project_id, commit_id)
    seeded_killed_by_unit_test = len(seeded_killed_by_unit_test)
    seeded_killed_by_integration_test = len(seeded_killed_by_integration_test)
    seeded_killed_by_both = len(seeded_killed_by_both)

    # Mutation Data Set
    mutation_killed_by_unit_test, mutation_killed_by_int_test, mutation_killed_by_both = \
        get_number_of_killed_mutations(mutations, set())

    # Disjoint Mutation Data Set
    disjoint_mutants = get_disjoint_mutant_set(mutations)
    disj_mutation_killed_by_unit_test, disj_mutation_killed_by_int_test, disj_mutation_killed_by_both = \
        get_number_of_killed_mutations(mutations, disjoint_mutants)

    df.loc[len(df)] = [project_name, 'SEEDED', 'Unit', seeded_killed_by_unit_test, seeded_killed_by_unit_test/(ut_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'SEEDED', 'Integration', seeded_killed_by_integration_test, seeded_killed_by_integration_test/(it_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'SEEDED', 'Both', seeded_killed_by_both, seeded_killed_by_both/((ut_testlc_sum+it_testlc_sum)/1000)]
    df.loc[len(df)] = [project_name, 'ALL', 'Unit', mutation_killed_by_unit_test, mutation_killed_by_unit_test/(ut_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'ALL', 'Integration', mutation_killed_by_int_test, mutation_killed_by_int_test/(it_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'ALL', 'Both', mutation_killed_by_both, mutation_killed_by_both/((ut_testlc_sum+it_testlc_sum)/1000)]
    df.loc[len(df)] = [project_name, 'DISJ', 'Unit', disj_mutation_killed_by_unit_test, disj_mutation_killed_by_unit_test/(ut_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'DISJ', 'Integration', disj_mutation_killed_by_int_test, disj_mutation_killed_by_int_test/(it_testlc_sum/1000)]
    df.loc[len(df)] = [project_name, 'DISJ', 'Both', disj_mutation_killed_by_both, disj_mutation_killed_by_both/((ut_testlc_sum+it_testlc_sum)/1000)]

    # Create Tables. This is easier than filtering everything out of the df again
    table_seeded.append([
        project_name, seeded_killed_by_unit_test, seeded_killed_by_integration_test, seeded_killed_by_both,
        "%.3f" % (seeded_killed_by_unit_test/(ut_testlc_sum/1000)),
        "%.3f" % (seeded_killed_by_integration_test/(it_testlc_sum/1000))
    ])

    table_mutation.append([
        project_name, mutation_killed_by_unit_test, mutation_killed_by_int_test, mutation_killed_by_both,
        "%.3f" % (mutation_killed_by_unit_test / (ut_testlc_sum / 1000)),
        "%.3f" % (mutation_killed_by_int_test / (it_testlc_sum / 1000))
    ])

    table_disj_mutation.append([
        project_name, disj_mutation_killed_by_unit_test, disj_mutation_killed_by_int_test, disj_mutation_killed_by_both,
        "%.3f" % (disj_mutation_killed_by_unit_test / (ut_testlc_sum / 1000)),
        "%.3f" % (disj_mutation_killed_by_int_test / (it_testlc_sum / 1000))
    ])

data_sets_with_table_handler = {
    'SEEDED': table_seeded,
    'ALL': table_mutation,
    'DISJ': table_disj_mutation
}

test_levels = ['Unit', 'Integration']

for data_set, table_handler in data_sets_with_table_handler.items():
    # Visualization
    sns.boxplot(x="Test Level", y="Score", data=df.loc[(df['Data Set'] == data_set) &
                                                       (df['Test Level'].isin(test_levels))]).set_title("Data Set: "+data_set)
    sns.swarmplot(x="Test Level", y="Score", data=df.loc[(df['Data Set'] == data_set) &
                                                         (df['Test Level'].isin(test_levels))], color=".25")
    plt.savefig(os.path.join(os.path.dirname(__file__), "figures", "%s_overall.pdf" % data_set))
    plt.show()


    # Table
    print("Table for Data Set: %s" % data_set)
    table_handler.append(['Mean',
                         "%.3f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Killed']),
                         "%.3f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Killed']),
                         "%.3f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Both'), 'Killed']),
                         "%.3f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Score']),
                         "%.3f" % statistics.mean(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Score'])])
    table_handler.append(['StDev',
                         "%.3f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Killed']),
                         "%.3f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Killed']),
                         "%.3f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Both'), 'Killed']),
                         "%.3f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Score']),
                         "%.3f" % statistics.stdev(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Score'])])
    print(tabulate(table_handler, tablefmt="latex"))

    # Statistical Tests
    print("Statistical Tests for data set %s" % data_set)
    print(stats.shapiro(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Score']))
    print(stats.shapiro(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Score']))
    print(stats.levene(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Score'],
                       df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Score']))
    print(stats.mannwhitneyu(df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Unit'), 'Score'],
                             df.loc[(df['Data Set'] == data_set) & (df['Test Level'] == 'Integration'), 'Score']))

elapsed = timeit.default_timer() - start_time
print("Execution time: %0.5f s" % elapsed)
